from datetime import datetime
from pathlib import Path
from typing import Any, Dict, List, Optional, Union

from git import InvalidGitRepositoryError
from git.repo import Repo
from loguru import logger
from pydantic import BaseModel, validator

from trubrics.ui.auth import expire_after_n_seconds, get_trubrics_auth_token
from trubrics.ui.firestore import add_document_to_project_subcollection
from trubrics.ui.trubrics_config import load_trubrics_config


class Feedback(BaseModel):
    """
    Dataclass for feedback given by a user from a UI component.

    Attributes:
        feedback_type: feedback type ['issue', 'faces', 'thumbs', 'custom']
        user_response: dict of all feedback given by the user
        data: a reference to the dataset used in the app
        model: a reference to the model used in the app
        collaborators: users who have collaborated so far on the issue
        open: whether the feedback item is open or closed
        tags: list of tags for the feedback issue
        git_commit: a git commit hash from the git repo where the app was run to collect feedback
        timestamp: timestamp at which the feedback was recorded
        created_by: who the feedback was run by
        closed_on: timestamp when the feedback was closed
        closed_by: who the feedback was run by
        metadata: free textual metadata field
    """

    feedback_type: str
    user_response: Dict[str, Union[float, int, str, bool]]
    data: Optional[str]
    model: Optional[str] = None
    collaborators: List[Optional[str]] = []
    open: bool = True
    tags: Optional[List[str]] = None
    git_commit: Optional[str] = None
    timestamp: Optional[int] = None
    created_by: Optional[Dict[str, Optional[str]]] = None
    closed_on: Optional[str] = None
    closed_by: Optional[str] = None
    metadata: Optional[Dict[str, Union[List[Any], float, int, str, dict]]] = None

    @validator("feedback_type")
    def target_column_must_be_in_data(cls, v):
        if v not in ["issue", "faces", "thumbs", "custom"]:
            raise ValueError("feedback_type must be one of ['issue', 'faces', 'thumbs', 'custom'].")
        return v

    def save_local(self, path: Optional[str] = None):
        self._set_fields_on_save()
        if path is None:
            path = f"./{self.timestamp}_feedback.json"
        with open(Path(path).absolute(), "w") as file:
            file.write(self.json(indent=4))
            logger.info(f"Feedback saved to {path}.")

    def save_ui(self, email: Optional[str], password: Optional[str]):
        trubrics_config = load_trubrics_config()
        self._set_fields_on_save()

        if email is None and password is None:
            email = trubrics_config.email
            password = trubrics_config.password.get_secret_value()

        auth = get_trubrics_auth_token(
            trubrics_config.firebase_auth_api_url, email, password, rerun=expire_after_n_seconds()
        )
        if "error" in auth:
            error_msg = f"Error in pushing feedback issue with email '{email}' to the Trubrics UI: {auth['error']}"
            logger.error(error_msg)
            raise Exception(error_msg)
        else:
            self.created_by = {"email": auth["email"], "displayName": auth["displayName"]}
            self.collaborators.append(auth["displayName"])

            res = add_document_to_project_subcollection(
                auth,
                firestore_api_url=trubrics_config.firestore_api_url,
                project=trubrics_config.project,
                subcollection="feedback",
                document_id=self.timestamp,
                document_json=self.json(),
            )
            if "error" in res:
                error_msg = f"Error in pushing feedback issue to the Trubrics UI: {res}"
                logger.error(error_msg)
                raise Exception(error_msg)
            else:
                logger.info("Feedback issue saved to the Trubrics UI.")

    def _set_fields_on_save(self):
        self.timestamp = int(datetime.now().timestamp())
        try:
            self.git_commit = Repo(search_parent_directories=True).head.object.hexsha
        except InvalidGitRepositoryError:
            logger.warning(
                "Current directory is not a git repository. Run `trubrics run` inside a git repository to save the"
                " commit hash."
            )



import os
import subprocess
from pathlib import Path
from typing import Optional

import typer
from loguru import logger
from rich import print as rprint
from rich.progress import Progress, SpinnerColumn, TextColumn

from trubrics.cli.run import validate_trubric_run_context
from trubrics.ui.auth import get_trubrics_auth_token, get_trubrics_firebase_auth_api_url
from trubrics.ui.firestore import (
    get_trubrics_firestore_api_url,
    list_projects_in_organisation,
)
from trubrics.ui.trubrics_config import (
    TrubricsConfig,
    TrubricsDefaults,
    load_trubrics_config,
)

app = typer.Typer(pretty_exceptions_show_locals=False)


def version_callback(value: bool):
    if value:
        import trubrics

        typer.echo(trubrics.__version__)
        raise typer.Exit()


@app.callback()
def main(
    version: bool = typer.Option(None, "--version", callback=version_callback, is_eager=True),
):
    return None


@app.command()
def init(
    firebase_api_key: Optional[str] = None,
    firebase_project_id: Optional[str] = None,
    trubrics_user: bool = typer.Option(False, prompt="Do you already have an account with Trubrics?"),
    project_name: Optional[str] = None,
):
    """
    `trubrics init` initialises an environment and authenticates with a Trubrics platform account.
    It does this by storing a `~/.trubrics_config.json` file in the root directory.
    Be guided through the process by answering the prompts, or feed these in programmatically with
    the environment variables `TRUBRICS_CONFIG_EMAIL` and `TRUBRICS_CONFIG_PASSWORD`:

    ```bash
    export TRUBRICS_CONFIG_EMAIL=<your_trubrics_email>
    ```
    ```bash
    export TRUBRICS_CONFIG_PASSWORD=<your_trubrics_password>
    ```
    ```bash
    trubrics init --trubrics-user --project-name <your_project_name>
    ```

    Args:
        firebase_api_key: optional firebase api key
        firebase_project_id: optional firebase project ID
        trubrics_user: boolean of whether the user has a Trubrics account
        project_name: optional project name to initialise
    """
    if firebase_api_key or firebase_project_id:
        if firebase_api_key and firebase_project_id:
            defaults = TrubricsDefaults(firebase_api_key=firebase_api_key, firebase_project_id=firebase_project_id)
        else:
            raise ValueError("Both API key and firebase_project_id are required to change project.")
    else:
        defaults = TrubricsDefaults()

    if trubrics_user:
        email = os.environ.get("TRUBRICS_CONFIG_EMAIL")
        password = os.environ.get("TRUBRICS_CONFIG_PASSWORD")
        if email or password:
            if not email or not password:
                logger.warning(
                    f"{'TRUBRICS_CONFIG_EMAIL' if not email else 'TRUBRICS_CONFIG_PASSWORD'} environment variable is"
                    " not set, asking user for a prompt."
                )
        email = email or typer.prompt("Enter your user email")
        password = password or typer.prompt("Enter your user password", hide_input=True)

        with Progress(
            SpinnerColumn(),
            TextColumn("[progress.description]{task.description}"),
            transient=True,
        ) as progress:
            progress.add_task(description="Authenticating user...", total=None)
            firebase_auth_api_url = get_trubrics_firebase_auth_api_url(defaults.firebase_api_key)
            auth = get_trubrics_auth_token(firebase_auth_api_url, email, password)
            if "error" in auth:
                rprint(f"Error in login email '{email}' to the Trubrics UI: {auth['error']}")
                raise typer.Abort()
            else:
                firestore_api_url = get_trubrics_firestore_api_url(auth, defaults.firebase_project_id)

            projects = list_projects_in_organisation(firestore_api_url=firestore_api_url, auth=auth)

            rprint(f"\n[bold yellow]Welcome {auth['displayName']}[bold yellow] :sunglasses:\n")

        if project_name:
            if project_name not in projects:
                raise ValueError(
                    f"Project '{project_name}' not found in organisation '{firestore_api_url.split('/')[-1]}'."
                    f" Projects currently available: {projects}."
                )
        else:
            if len(projects) > 0:
                for index, project in enumerate(projects):
                    rprint(f"[bold green][{index}][/bold green] [green]{project}[/green]")
                project_num = typer.prompt("Select your project (e.g. 0)")

                project_int = int(project_num)
                if project_int not in list(range(len(projects))):
                    message = typer.style(
                        f"Project [{project_num}] not recognised."
                        "Please indicate an integer referring to one of the project names"
                        " above.",
                        fg=typer.colors.RED,
                        bold=True,
                    )
                    typer.echo(message)
                    raise typer.Abort()
                else:
                    project_name = projects[project_int]
            else:
                message = typer.style(
                    f"Organisation '{firestore_api_url.split('/')[-1]}' has no projects created."
                    " Navigate to the Trubrics UI to add a project.",
                    fg=typer.colors.RED,
                    bold=True,
                )
                typer.echo(message)
                raise typer.Abort()

        trubrics_config = TrubricsConfig(
            firebase_auth_api_url=firebase_auth_api_url,
            firestore_api_url=firestore_api_url,
            username=auth["displayName"],
            email=email,
            password=password,  # type: ignore
            project=project_name,  # type: ignore
        )
        trubrics_config.save()

        typer.echo(typer.style("Successful authentication with configuration:", fg=typer.colors.GREEN, bold=True))
        rprint(trubrics_config.dict())
        rprint()
        rprint(
            "[bold green]You can now push trubrics and feedback to the Trubrics platform:"
            f"\n{defaults.trubrics_url}[bold green]\n"
        )
    else:
        rprint(
            "[bold orange_red1]Sign up here to get access to the Trubrics platform:"
            f"\n\n{defaults.demo_sign_up_url}[bold orange_red1]\n"
        )


def _framework_callback(value: str):
    if value not in ["gradio", "streamlit", "dash"]:
        raise typer.BadParameter("Only 'gradio', 'dash' or 'streamlit' frameworks are supported.")
    return value


@app.command()
def example_app(
    framework: str = typer.Option("streamlit", callback=_framework_callback),
    trubrics_platform_auth: Optional[str] = None,
):
    """Runs the example user feedback collector app.

    Args:
        framework: framework of streamlit, dash or gradio
        trubrics_platform_auth: whether to save feedback to the Trubrics platform
    """
    dirname = os.path.dirname(__file__)
    filename = os.path.join(dirname, f"../example/app_titanic_{framework}.py")
    if trubrics_platform_auth not in [None, "single_user", "multiple_users"]:
        raise ValueError(
            f"trubrics_platform_auth={trubrics_platform_auth} not recognised. Must be one of [None, 'single_user',"
            " 'multiple_users']."
        )
    if framework == "streamlit":
        if trubrics_platform_auth:
            subprocess.call(["streamlit", "run", filename, "--", "--trubrics-platform-auth", trubrics_platform_auth])
        else:
            subprocess.call(["streamlit", "run", filename])
    elif framework in ["gradio", "dash"]:
        if trubrics_platform_auth:
            raise ValueError("Trubrics auth currently only available with Streamlit.")
        subprocess.call(["python3", filename])


@app.command()
def run(
    save_ui: bool = typer.Option(False, prompt="Would you like to save your trubric to the UI?"),
    run_context_path: str = typer.Option(
        default="example", prompt="Enter the path to your trubric run .py file. Press enter for an"
    ),
    trubric_output_file_path: str = "./my_new_trubric.json",
    raise_on_failure: bool = True,
):
    """Runs an example trubric (list of model validations) on the titanic dataset.

    Args:
        save_ui: whether to save validations to the UI with in app user authentication
        run_context_path: path to the trubrics run context
        trubric_output_file_path: path to save your output trubric file
        raise_on_failure: raise an exception on failure of trubric
    """
    trubric_run_path = None
    if run_context_path != "example":
        trubric_run_path = Path(run_context_path).absolute()
        if not trubric_run_path.exists():
            rprint(f"[red]Path '{trubric_run_path}' not found.[red]")
            raise typer.Abort()
        rc = validate_trubric_run_context(str(trubric_run_path)).RUN_CONTEXT
    else:
        from trubrics.example.trubric_run import RUN_CONTEXT as rc  # type: ignore
    rprint(
        f"\nRunning trubric from file '{trubric_run_path or 'trubrics.example.trubric_run'}' with model"
        f" '{rc.model_name}' and dataset '{rc.data_context.name}'.\n"
    )
    new_trubric = rc.set_new_trubric()
    if save_ui:
        trubrics_config = load_trubrics_config().dict()
        if trubrics_config["email"] is not None:
            new_trubric.save_ui(raise_on_failure=raise_on_failure)
        else:
            typer.echo(
                typer.style(
                    "ERROR: You must authenticate with the Trubrics platform by running `trubrics init` to remotely"
                    " save trubrics runs.",
                    fg=typer.colors.RED,
                )
            )
    else:
        defaults = TrubricsDefaults()
        new_trubric.save_local(trubric_output_file_path, raise_on_failure=raise_on_failure)
        rprint(
            "\n[bold orange_red1]Be sure to check out our docs to see how you can leverage the Trubrics platform."
            f"\n\n{defaults.demo_sign_up_url}[bold orange_red1]\n"
        )


if __name__ == "__main__":
    app()



from typing import Any, Dict, List, Optional, Tuple, Union

import streamlit as st

from trubrics.feedback import config
from trubrics.feedback.dataclass import Feedback
from trubrics.ui.auth import expire_after_n_seconds, get_trubrics_auth_token
from trubrics.ui.trubrics_config import load_trubrics_config


class FeedbackCollector:
    def __init__(
        self,
        data: Optional[str] = None,
        model: Optional[str] = None,
        trubrics_platform_auth: Optional[str] = None,
    ):
        """
        The FeedbackCollector allows Data Scientists to collect feedback from within their app.

        Args:
            data: a reference to the data that was used to collect the feedback (e.g. a link to a dataset)
            model: a reference to the model that was used to collect the feedback (e.g. a link to a model)
            trubrics_platform_auth: option to save the feedback to the trubrics platform

                - *None*: save feedback locally to .json
                - *single_user*: save feedback directly to the Trubrics platform,
                    using auth credentials of the app owner
                - *multiple_users*: save feedback directly to the Trubrics platform, with full user auth
        """
        self.data = data
        self.model = model
        if trubrics_platform_auth in [None, "single_user", "multiple_users"]:
            self.trubrics_platform_auth = trubrics_platform_auth
        else:
            raise ValueError(
                f"trubrics_platform_auth={trubrics_platform_auth} not recognised. Must be one of [None, 'single_user',"
                " 'multiple_users']."
            )

        self.email = ""
        self.password = ""
        self.authenticated = False

    def st_trubrics_auth(self):
        if self.trubrics_platform_auth is None:
            raise ValueError(
                "The `.st_trubrics_auth()` method is reserved for usage with the Trubrics platform. See the"
                " 'trubrics_platform_auth' argument for authentication options."
            )

        trubrics_config = load_trubrics_config()
        if self.authenticated:
            st.write(self.email + " signed in.")
            if st.button("Sign out"):
                self.authenticated = False
                self.st_trubrics_auth()
        else:
            with st.form("auth form"):
                self.email = st.text_input(
                    label=config.USER_EMAIL,
                    placeholder=config.USER_EMAIL,
                    label_visibility="collapsed",
                    key="email",
                )
                self.password = st.text_input(
                    label=config.USER_PASSWORD,
                    placeholder=config.USER_PASSWORD,
                    label_visibility="collapsed",
                    key="password",
                    type="password",
                )
                submitted = st.form_submit_button("Sign In")
                if submitted:
                    # check auth
                    auth = get_trubrics_auth_token(
                        trubrics_config.firebase_auth_api_url, self.email, self.password, rerun=expire_after_n_seconds()
                    )
                    if "error" in auth:
                        st.error(f"Error authenticating user {self.email}. Try again or contact your admin team.")
                    else:
                        st.success(f"{self.email} successfully signed in.")
                        self.authenticated = True

    def st_feedback(
        self,
        feedback_type: str = "issue",
        user_response: Optional[Dict[str, Union[float, int, str, bool]]] = None,
        path: Optional[str] = None,
        metadata: Optional[Dict[str, Any]] = None,
        tags: Optional[List[str]] = None,
        key: Optional[str] = None,
        open_feedback_label: Optional[str] = None,
    ) -> Optional[Feedback]:
        """
        Collect user feedback within a Streamlit web application.

        Args:
            feedback_type: type of feedback to be collected

                - issue: issue with a open text title and description fields
                - thumbs: positive or negative feedback with thumbs emojis
                - faces: a scale of 1 to 5 with face emojis
                - custom: a customisable feedback type that allows users to specify the user_response dict
            user_response: a dict of user responses to save with your feedback for feedback_type='custom'
            path: path to save feedback local .json. Defaults to "./*timestamp*_feedback.json"
            metadata: data to save with your feedback
            tags: a list of tags for your feedback
            key: a key for each streamlit component
            open_feedback_label: label of optional text_input for "faces" or "thumbs" feedback_type
        """
        if key is None:
            key = feedback_type
        if feedback_type == "issue":
            if user_response or open_feedback_label:
                raise ValueError("For feedback_type='issue', title, description and open_feedback_label must be None.")
            issue_data = self.st_issue_ui(key)
            if issue_data:
                return self._save_feedback(
                    feedback_type=feedback_type,
                    path=path,
                    metadata=metadata,
                    user_response={issue_data[0]: issue_data[1]},
                    tags=tags,
                )
        elif feedback_type in ("thumbs", "faces"):
            if user_response:
                raise ValueError(
                    f"For feedback_type='{feedback_type}', user_response is set inside the component (must be None)."
                )
            return self._save_quantitative_feedback(
                feedback_type=feedback_type,
                path=path,
                metadata=metadata,
                tags=tags,
                key=key,
                open_feedback_label=open_feedback_label,
            )
        elif feedback_type == "custom":
            if user_response:
                return self._save_feedback(
                    feedback_type=feedback_type,
                    user_response=user_response,
                    path=path,
                    metadata=metadata,
                    tags=tags,
                )
            else:
                raise ValueError("For feedback_type='custom', title and description parameters must be specified.")
        else:
            raise ValueError("feedback_type must be one of ['issue', 'faces', 'thumbs', 'custom'].")
        return None

    def _save_feedback(
        self,
        feedback_type: str,
        user_response: Dict[str, Union[float, int, str, bool]],
        path: Optional[str] = None,
        metadata: Optional[Dict[str, Any]] = None,
        tags: Optional[List[str]] = None,
    ) -> Optional[Feedback]:
        feedback = Feedback(
            feedback_type=feedback_type,
            user_response=user_response,
            data=self.data,
            model=self.model,
            metadata=metadata,
            tags=tags,
        )
        if self.trubrics_platform_auth is None:
            feedback.save_local(path=path)
            st.success(config.LOCAL_SAVE)
        elif self.trubrics_platform_auth == "multiple_users":
            if self.authenticated:
                feedback.save_ui(self.email, self.password)
                st.success(config.PLATFORM_SAVE)
            else:
                st.error("User is not authenticated. Please try again or contact your admin.")
        elif self.trubrics_platform_auth == "single_user":
            feedback.save_ui(None, None)
            st.success(config.PLATFORM_SAVE)
        else:
            raise ValueError(
                f"trubrics_platform_auth={self.trubrics_platform_auth} not recognised. Must be one of [None,"
                " 'single_user', 'multiple_users']."
            )
        return feedback

    def _save_quantitative_feedback(
        self,
        feedback_type,
        key,
        open_feedback_label: Optional[str],
        path: Optional[str] = None,
        metadata: Optional[Dict[str, Any]] = None,
        tags: Optional[List[str]] = None,
    ) -> Optional[Feedback]:
        if f"{key}_state" not in st.session_state:
            st.session_state[f"{key}_state"] = ""
        if f"{key}_save_button" not in st.session_state:
            st.session_state[f"{key}_save_button"] = False
        if f"previous_{key}_state" not in st.session_state:
            st.session_state[f"previous_{key}_state"] = ""

        title = f"User satisfaction: {feedback_type}"

        def feedback_handler(open_feedback_label):
            user_response = {
                title: st.session_state[f"{key}_state"],
                open_feedback_label: st.session_state[f"{feedback_type}_open_feedback"].rstrip(),
            }
            st.session_state[f"previous_{key}_state"] = user_response
            st.session_state[f"{key}_state"] = ""

            self._enable_all_buttons(feedback_type=feedback_type)

        ui_state = getattr(self, f"st_{feedback_type}_ui")(
            key=key, disable_on_click=True if open_feedback_label else False
        )
        if ui_state:
            st.session_state[f"{key}_state"] = ui_state

        if open_feedback_label:
            if st.session_state.get(f"{key}_state_clicked", "") != "":
                st.text_input(open_feedback_label, key=f"{feedback_type}_open_feedback")
                st.button(
                    config.FEEDBACK_SAVE_BUTTON,
                    on_click=feedback_handler,
                    args=(open_feedback_label,),
                    key=f"{key}_save_button",
                )
        else:
            if ui_state:
                user_response = {title: ui_state}
                return self._save_feedback(
                    user_response=user_response, feedback_type=feedback_type, metadata=metadata, path=path, tags=tags
                )
        if st.session_state[f"{key}_save_button"]:
            user_response = {title: ui_state}
            return self._save_feedback(
                user_response=st.session_state[f"previous_{key}_state"],
                feedback_type=feedback_type,
                metadata=metadata,
                path=path,
                tags=tags,
            )
        return None

    @staticmethod
    def st_issue_ui(key: Optional[str] = None) -> Optional[Tuple[str, str]]:
        if key is None:
            key = "issue"

        if f"{key}_save_button" not in st.session_state:
            st.session_state[f"{key}_save_button"] = False

        if f"previous_{key}_state" not in st.session_state:
            st.session_state[f"previous_{key}_state"] = ""

        def clear_session_state():
            st.session_state[f"previous_{key}_state"] = (
                st.session_state[f"{key}_title"],
                st.session_state[f"{key}_description"],
            )
            st.session_state[f"{key}_title"] = ""
            st.session_state[f"{key}_description"] = ""

        title = st.text_input(
            label=config.TITLE,
            help=config.TITLE_EXPLAIN,
            key=f"{key}_title",
        )
        description = st.text_input(
            label=config.DESCRIPTION,
            help=config.DESCRIPTION_EXPLAIN,
            key=f"{key}_description",
        )
        enabled = title and description
        if enabled:
            st.button(config.FEEDBACK_SAVE_BUTTON, on_click=clear_session_state, key=f"{key}_save_button")
        if st.session_state[f"{key}_save_button"]:
            return st.session_state[f"previous_{key}_state"]
        else:
            return None

    def st_thumbs_ui(self, disable_on_click: bool = False, key: Optional[str] = None) -> Optional[str]:
        if key is None:
            key = "thumbs"

        button_states = [f"{key}_1", f"{key}_2"]
        col1, col2 = st.columns([1, 15])
        with col1:
            up = self._emoji_button("👍", key, disable_on_click, button_states, 1)
        with col2:
            down = self._emoji_button("👎", key, disable_on_click, button_states, 2)

        if up:
            return self._return_quantitative_ui_button(key, disable_on_click, button_states, 1, "thumbs up")
        elif down:
            return self._return_quantitative_ui_button(key, disable_on_click, button_states, 2, "thumbs down")
        else:
            return None

    def st_faces_ui(self, disable_on_click: bool = False, key: Optional[str] = None) -> Optional[str]:
        if key is None:
            key = "faces"

        button_states = [f"{key}_1", f"{key}_2", f"{key}_3", f"{key}_4", f"{key}_5"]
        col1, col2, col3, col4, col5 = st.columns([1, 1, 1, 1, 10])
        with col1:
            one = self._emoji_button("😞", key, disable_on_click, button_states, 1)
        with col2:
            two = self._emoji_button("🙁", key, disable_on_click, button_states, 2)
        with col3:
            three = self._emoji_button("😐", key, disable_on_click, button_states, 3)
        with col4:
            four = self._emoji_button("🙂", key, disable_on_click, button_states, 4)
        with col5:
            five = self._emoji_button("😀", key, disable_on_click, button_states, 5)

        if one:
            return self._return_quantitative_ui_button(key, disable_on_click, button_states, 1, "very negative")
        elif two:
            return self._return_quantitative_ui_button(key, disable_on_click, button_states, 2, "negative")
        elif three:
            return self._return_quantitative_ui_button(key, disable_on_click, button_states, 3, "neutral")
        elif four:
            return self._return_quantitative_ui_button(key, disable_on_click, button_states, 4, "positive")
        elif five:
            return self._return_quantitative_ui_button(key, disable_on_click, button_states, 5, "very positive")
        else:
            return None

    @staticmethod
    def _return_quantitative_ui_button(ui_type, disable_on_click, button_states, index, output_str):
        if disable_on_click:
            if st.session_state[f"{ui_type}_state_clicked"] == button_states[index - 1]:
                return f":{index} - {output_str}:"
            else:
                return None
        else:
            return f":{index} - {output_str}:"

    def _emoji_button(self, emoji, key, disable_on_click, button_states, index):
        return st.button(
            emoji,
            key=f"{key}_{index}",
            on_click=self._disable_buttons,
            args=(key, disable_on_click, index, button_states),
            disabled=st.session_state.get(f"{key}_{index}_disable", False),
        )

    @staticmethod
    def _disable_buttons(key, disable_on_click, index, button_states):
        """Disable all buttons except the one that was clicked, and re-enable all buttons if re-clicked."""
        if disable_on_click:
            if any([st.session_state.get(button_state + "_disable", False) for button_state in button_states]):
                st.session_state[f"{key}_state_clicked"] = ""
                for button_state in button_states:
                    st.session_state[button_state + "_disable"] = False
            else:
                enabled = button_states.pop(index - 1)
                st.session_state[enabled + "_disable"] = False
                st.session_state[f"{key}_state_clicked"] = enabled
                for button_state in button_states:
                    st.session_state[button_state + "_disable"] = True

    @staticmethod
    def _enable_all_buttons(feedback_type):
        for key in st.session_state:
            if key.endswith("_disable") and key.startswith(feedback_type):
                st.session_state[key] = False
            if key == f"{feedback_type}_state_clicked":
                st.session_state[key] = ""



import json
import time
from functools import lru_cache
from typing import Dict

import requests  # type: ignore


def get_trubrics_firebase_auth_api_url(firebase_api_key):
    return f"https://identitytoolkit.googleapis.com/v1/accounts:signInWithPassword?key={firebase_api_key}"


def expire_after_n_seconds(seconds=600):
    """Return the same value within `seconds` time period."""
    return round(time.time() / seconds)


@lru_cache(maxsize=32)
def get_trubrics_auth_token(firebase_auth_api_url, email, password, rerun=None) -> Dict[str, str]:
    del rerun  # this variable is just used to force a refresh of lru_cache
    try:
        r = requests.post(
            firebase_auth_api_url,
            headers={"Content-Type": "application/json"},
            data=json.dumps({"email": email, "password": password, "returnSecureToken": True}),
            timeout=5000,
        )
        r.raise_for_status()
        auth_response = json.loads(r.text)

        return {
            "idToken": auth_response["idToken"],
            "email": auth_response["email"],
            "uid": auth_response["localId"],
            "displayName": auth_response["displayName"],
        }
    except requests.exceptions.RequestException as err:
        return {"error": str(err)}



import gradio as gr

from trubrics.integrations.gradio import collect_feedback

with gr.Blocks() as demo:
    gr.Markdown("Gradio app to collect user feedback on models.")
    with gr.Tab("Feedback"):
        collect_feedback(tags=["Gradio"])

demo.launch()



import dash_bootstrap_components as dbc
from dash import Dash, html

from trubrics.integrations.dash import collect_feedback

app = Dash(__name__, external_stylesheets=[dbc.themes.BOOTSTRAP])

app.layout = html.Div([collect_feedback(tags=["Dash"], save_ui=False)])

if __name__ == "__main__":
    app.run_server(debug=True)



from typing import Optional

import streamlit as st
import typer

from trubrics.context import DataContext
from trubrics.example import get_titanic_data_and_model
from trubrics.example import titanic_config as tc
from trubrics.integrations.streamlit import (
    FeedbackCollector,
    generate_what_if_streamlit,
)

cli = typer.Typer()


@st.cache(allow_output_mutation=True)
def init_trubrics(trubrics_platform_auth):
    _, test_df, model = get_titanic_data_and_model()

    data_context = DataContext(
        testing_data=test_df,
        target="Survived",
        categorical_columns=tc.CATEGORICAL_COLUMNS,
        business_columns=tc.BUSINESS_COLUMNS,
    )

    collector = FeedbackCollector(
        data="trubrics data_context",
        model="rf_model",
        trubrics_platform_auth=trubrics_platform_auth,
    )
    return model, data_context, collector


def feedback_example(feedback_type, collector, metadata, open_feedback_label=None, user_response=None):
    file_name = f"{feedback_type}_feedback.json"
    if user_response:
        feedback = collector.st_feedback(
            feedback_type=feedback_type, metadata=metadata, path=file_name, user_response=user_response
        )
    else:
        code_snippet = f"""
        from trubrics.integrations.streamlit import FeedbackCollector
collector = FeedbackCollector()
collector.st_feedback(
    feedback_type="{feedback_type}"{f', open_feedback="{open_feedback_label}"' if open_feedback_label else ''}
)
        """
        feedback = collector.st_feedback(
            feedback_type=feedback_type, metadata=metadata, path=file_name, open_feedback_label=open_feedback_label
        )
        with st.expander(f"See code snippet for feedback_type='{feedback_type}'"):
            st.code(code_snippet)
    if feedback:
        st.write("Example of the FeedbackCollector's output:")
        st.write(feedback.dict())
        st.download_button("Download this example .json file", feedback.json(), mime="text/json", file_name=file_name)
        st.markdown(
            """
            As you collect feedback, .json files are saved to either you local filesystem, or to the Trubrics platform
            (see [here](https://trubrics.github.io/trubrics-sdk/feedback/) for more info).
            """
        )
    st.markdown("***")


@cli.command()
def main(trubrics_platform_auth: Optional[str] = None):
    model, data_context, collector = init_trubrics(trubrics_platform_auth)

    st.title("Titanic Demo App")
    with st.expander("Notes on the demo"):
        st.markdown(
            """
        The sidebar of features and the '**Model prediction**' section below are not a part
        of the FeedbackCollector, but rather a demo of a basic application of how
        you could allow users to test your model and provide feedback.
        """
        )

    with st.sidebar:
        if trubrics_platform_auth == "multiple_users":
            st.title("Authentication")
            collector.st_trubrics_auth()
        st.title("Test the model with different inputs")
        df = generate_what_if_streamlit(data_context=data_context)
    wi_prediction = model.predict(df)[0]

    metadata = {"data": df.to_dict(), "prediction": wi_prediction}

    prediction = "## " + str(wi_prediction) + f"{' (survived)' if wi_prediction else  ' (did not survive)'}"
    col1, col2 = st.columns(2)
    with col1:
        st.markdown("## Model prediction:")
    with col2:
        st.markdown(prediction, unsafe_allow_html=True)

    st.markdown("")
    st.markdown("")
    st.markdown("## FeedbackCollector Examples")
    st.markdown(
        """
        Here are examples of how you could implement different types of feedback within your app. Each
        feedback component is:
        - Independent, and can be duplicated around your app
        - Able to collect different data from a specific part of your app
        - Saved to your local filesystem, with a customisable path
        """
    )
    st.markdown("***")

    st.markdown('##### 1 - "Does this prediction look correct?"')
    thumbs_open_feedback = st.radio(
        "Add open feedback?",
        ("No open feedback", "With open feedback"),
        label_visibility="collapsed",
        key="thumbs_radio",
        horizontal=True,
    )
    if thumbs_open_feedback == "With open feedback":
        feedback_example(
            "thumbs", collector=collector, metadata=metadata, open_feedback_label="Please provide a description"
        )
    elif thumbs_open_feedback == "No open feedback":
        feedback_example("thumbs", collector=collector, metadata=metadata)
    else:
        raise NotImplementedError()

    st.markdown('##### 2 - "How satisfied are you with this prediction?"')
    faces_open_feedback = st.radio(
        "Add open feedback?",
        ("No open feedback", "With open feedback"),
        label_visibility="collapsed",
        key="faces_radio",
        horizontal=True,
    )
    if faces_open_feedback == "With open feedback":
        feedback_example(
            "faces", collector=collector, metadata=metadata, open_feedback_label="Please provide a description"
        )
    elif faces_open_feedback == "No open feedback":
        feedback_example("faces", collector=collector, metadata=metadata)
    else:
        raise NotImplementedError()

    st.markdown('##### 3 - "Raise a specific issue"')
    feedback_example("issue", collector=collector, metadata=metadata)

    custom_question = "How much do you love this component?"
    st.markdown(f'##### 4 - "{custom_question}"')
    slider = st.slider("Custom feedback slider", max_value=10, value=9)
    submit = st.button("Save feedback")
    code_snippet = """
    from trubrics.integrations.streamlit import FeedbackCollector
import streamlit as st

collector = FeedbackCollector()

slider = st.slider("Custom feedback slider", max_value=10, value=9)
submit = st.button("Save feedback")

if submit and slider:
    collector.st_feedback(
        "custom",
        user_response={
            "How much do you love this component?": slider,
        }
    )
        """
    with st.expander("See code snippet for feedback_type='custom'"):
        st.code(code_snippet)
    if submit and slider:
        feedback_example(
            "custom",
            collector=collector,
            metadata=metadata,
            user_response={
                custom_question: slider,
            },
        )


if __name__ == "__main__":
    cli(standalone_mode=False)
